"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("util");
const http = require("http");
const https = require("https");
const cacheable_lookup_1 = require("cacheable-lookup");
const CacheableRequest = require("cacheable-request");
const is_1 = require("@sindresorhus/is");
const lowercaseKeys = require("lowercase-keys");
const toReadableStream = require("to-readable-stream");
const options_to_url_1 = require("./utils/options-to-url");
const errors_1 = require("./errors");
const merge_1 = require("./utils/merge");
const known_hook_events_1 = require("./known-hook-events");
const dynamic_require_1 = require("./utils/dynamic-require");
const get_body_size_1 = require("./utils/get-body-size");
const is_form_data_1 = require("./utils/is-form-data");
const supports_brotli_1 = require("./utils/supports-brotli");
const nonEnumerableProperties = [
    'context',
    'body',
    'json',
    'form'
];
exports.preNormalizeArguments = (options, defaults) => {
    var _a;
    // `options.headers`
    if (is_1.default.undefined(options.headers)) {
        options.headers = {};
    }
    else {
        options.headers = lowercaseKeys(options.headers);
    }
    // `options.prefixUrl`
    if (is_1.default.urlInstance(options.prefixUrl) || is_1.default.string(options.prefixUrl)) {
        options.prefixUrl = options.prefixUrl.toString();
        if (options.prefixUrl.length !== 0 && !options.prefixUrl.endsWith('/')) {
            options.prefixUrl += '/';
        }
    }
    else {
        options.prefixUrl = defaults ? defaults.prefixUrl : '';
    }
    // `options.hooks`
    if (is_1.default.undefined(options.hooks)) {
        options.hooks = {};
    }
    if (is_1.default.object(options.hooks)) {
        for (const event of known_hook_events_1.default) {
            if (Reflect.has(options.hooks, event)) {
                if (!is_1.default.array(options.hooks[event])) {
                    throw new TypeError(`Parameter \`${event}\` must be an Array, not ${is_1.default(options.hooks[event])}`);
                }
            }
            else {
                options.hooks[event] = [];
            }
        }
    }
    else {
        throw new TypeError(`Parameter \`hooks\` must be an Object, not ${is_1.default(options.hooks)}`);
    }
    if (defaults) {
        for (const event of known_hook_events_1.default) {
            // @ts-ignore TS is dumb.
            options.hooks[event] = [
                ...defaults.hooks[event],
                ...options.hooks[event]
            ];
        }
    }
    // `options.timeout`
    if (is_1.default.number(options.timeout)) {
        options.timeout = { request: options.timeout };
    }
    else if (!is_1.default.object(options.timeout)) {
        options.timeout = {};
    }
    // `options.retry`
    const { retry } = options;
    if (defaults) {
        options.retry = { ...defaults.retry };
    }
    else {
        options.retry = {
            calculateDelay: retryObject => retryObject.computedValue,
            limit: 0,
            methods: [],
            statusCodes: [],
            errorCodes: [],
            maxRetryAfter: undefined
        };
    }
    if (is_1.default.object(retry)) {
        options.retry = {
            ...options.retry,
            ...retry
        };
    }
    else if (is_1.default.number(retry)) {
        options.retry.limit = retry;
    }
    if (options.retry.maxRetryAfter === undefined) {
        options.retry.maxRetryAfter = Math.min(...[options.timeout.request, options.timeout.connect].filter(n => !is_1.default.nullOrUndefined(n)));
    }
    options.retry.methods = [...new Set(options.retry.methods.map(method => method.toUpperCase()))];
    options.retry.statusCodes = [...new Set(options.retry.statusCodes)];
    options.retry.errorCodes = [...new Set(options.retry.errorCodes)];
    // `options.dnsCache`
    if (options.dnsCache && !(options.dnsCache instanceof cacheable_lookup_1.default)) {
        options.dnsCache = new cacheable_lookup_1.default({ cacheAdapter: options.dnsCache });
    }
    // `options.method`
    if (is_1.default.string(options.method)) {
        options.method = options.method.toUpperCase();
    }
    else {
        options.method = ((_a = defaults) === null || _a === void 0 ? void 0 : _a.method) || 'GET';
    }
    // Better memory management, so we don't have to generate a new object every time
    if (options.cache) {
        options.cacheableRequest = new CacheableRequest((options, handler) => options.request(options, handler), options.cache);
    }
    // `options.cookieJar`
    if (is_1.default.object(options.cookieJar)) {
        let { setCookie, getCookieString } = options.cookieJar;
        // Horrible `tough-cookie` check
        if (setCookie.length === 4 && getCookieString.length === 0) {
            if (!Reflect.has(setCookie, util_1.promisify.custom)) {
                setCookie = util_1.promisify(setCookie.bind(options.cookieJar));
                getCookieString = util_1.promisify(getCookieString.bind(options.cookieJar));
            }
        }
        else if (setCookie.length !== 2) {
            throw new TypeError('`options.cookieJar.setCookie` needs to be an async function with 2 arguments');
        }
        else if (getCookieString.length !== 1) {
            throw new TypeError('`options.cookieJar.getCookieString` needs to be an async function with 1 argument');
        }
        options.cookieJar = { setCookie, getCookieString };
    }
    return options;
};
exports.mergeOptions = (...sources) => {
    const mergedOptions = exports.preNormalizeArguments({});
    // Non enumerable properties shall not be merged
    const properties = {};
    for (const source of sources) {
        if (!source) {
            continue;
        }
        merge_1.default(mergedOptions, exports.preNormalizeArguments(merge_1.default({}, source), mergedOptions));
        for (const name of nonEnumerableProperties) {
            if (!Reflect.has(source, name)) {
                continue;
            }
            properties[name] = {
                writable: true,
                configurable: true,
                enumerable: false,
                value: source[name]
            };
        }
    }
    Object.defineProperties(mergedOptions, properties);
    return mergedOptions;
};
exports.normalizeArguments = (url, options, defaults) => {
    // Merge options
    if (typeof url === 'undefined') {
        throw new TypeError('Missing `url` argument');
    }
    if (typeof options === 'undefined') {
        options = {};
    }
    if (is_1.default.urlInstance(url) || is_1.default.string(url)) {
        options.url = url;
        options = exports.mergeOptions(defaults && defaults.options, options);
    }
    else {
        if (Reflect.has(url, 'resolve')) {
            throw new Error('The legacy `url.Url` is deprecated. Use `URL` instead.');
        }
        options = exports.mergeOptions(defaults && defaults.options, url, options);
    }
    // Normalize URL
    // TODO: drop `optionsToUrl` in Got 12
    if (is_1.default.string(options.url)) {
        options.url = options.prefixUrl + options.url;
        options.url = options.url.replace(/^unix:/, 'http://$&');
        if (options.searchParams || options.search) {
            options.url = options.url.split('?')[0];
        }
        options.url = options_to_url_1.default({
            origin: options.url,
            ...options
        });
    }
    else if (!is_1.default.urlInstance(options.url)) {
        options.url = options_to_url_1.default({ origin: options.prefixUrl, ...options });
    }
    const normalizedOptions = options;
    // Make it possible to change `options.prefixUrl`
    let prefixUrl = options.prefixUrl;
    Object.defineProperty(options, 'prefixUrl', {
        set: (value) => {
            if (!normalizedOptions.url.href.startsWith(value)) {
                throw new Error(`Cannot change \`prefixUrl\` from ${prefixUrl} to ${value}: ${normalizedOptions.url.href}`);
            }
            normalizedOptions.url = new URL(value + normalizedOptions.url.href.slice(prefixUrl.length));
            prefixUrl = value;
        },
        get: () => prefixUrl
    });
    // Make it possible to remove default headers
    for (const [key, value] of Object.entries(options.headers)) {
        if (is_1.default.undefined(value)) {
            delete options.headers[key];
        }
        else if (is_1.default.null_(value)) {
            throw new TypeError('Use `undefined` instead of `null` to delete HTTP headers');
        }
    }
    for (const hook of options.hooks.init) {
        if (is_1.default.asyncFunction(hook)) {
            throw new TypeError('The `init` hook must be a synchronous function');
        }
        // @ts-ignore TS is dumb.
        hook(normalizedOptions);
    }
    return normalizedOptions;
};
const withoutBody = new Set(['GET', 'HEAD']);
exports.normalizeRequestArguments = async (options) => {
    var _a, _b;
    options = exports.mergeOptions(options);
    let uploadBodySize;
    // Serialize body
    const { headers } = options;
    const isForm = !is_1.default.undefined(options.form);
    const isJSON = !is_1.default.undefined(options.json);
    const isBody = !is_1.default.undefined(options.body);
    if ((isBody || isForm || isJSON) && withoutBody.has(options.method)) {
        throw new TypeError(`The \`${options.method}\` method cannot be used with a body`);
    }
    if ([isBody, isForm, isJSON].filter(isTrue => isTrue).length > 1) {
        throw new TypeError('The `body`, `json` and `form` options are mutually exclusive');
    }
    if (isBody) {
        if (is_1.default.object(options.body) && is_form_data_1.default(options.body)) {
            // Special case for https://github.com/form-data/form-data
            if (!Reflect.has(headers, 'content-type')) {
                headers['content-type'] = `multipart/form-data; boundary=${options.body.getBoundary()}`;
            }
        }
        else if (!is_1.default.nodeStream(options.body) && !is_1.default.string(options.body) && !is_1.default.buffer(options.body)) {
            throw new TypeError('The `body` option must be a stream.Readable, string or Buffer');
        }
    }
    else if (isForm) {
        if (!is_1.default.object(options.form)) {
            throw new TypeError('The `form` option must be an Object');
        }
        if (!Reflect.has(headers, 'content-type')) {
            headers['content-type'] = 'application/x-www-form-urlencoded';
        }
        options.body = (new URLSearchParams(options.form)).toString();
    }
    else if (isJSON) {
        if (!Reflect.has(headers, 'content-type')) {
            headers['content-type'] = 'application/json';
        }
        options.body = JSON.stringify(options.json);
    }
    // Convert buffer to stream to receive upload progress events (#322)
    if (is_1.default.buffer(options.body)) {
        uploadBodySize = options.body.length;
        options.body = toReadableStream(options.body);
    }
    else {
        uploadBodySize = await get_body_size_1.default(options);
    }
    // See https://tools.ietf.org/html/rfc7230#section-3.3.2
    // A user agent SHOULD send a Content-Length in a request message when
    // no Transfer-Encoding is sent and the request method defines a meaning
    // for an enclosed payload body.  For example, a Content-Length header
    // field is normally sent in a POST request even when the value is 0
    // (indicating an empty payload body).  A user agent SHOULD NOT send a
    // Content-Length header field when the request message does not contain
    // a payload body and the method semantics do not anticipate such a
    // body.
    if (!Reflect.has(headers, 'content-length') && !Reflect.has(headers, 'transfer-encoding')) {
        if ((options.method === 'POST' || options.method === 'PUT' || options.method === 'PATCH') &&
            !is_1.default.undefined(uploadBodySize)) {
            headers['content-length'] = String(uploadBodySize);
        }
    }
    if (!options.isStream && options.responseType === 'json' && is_1.default.undefined(headers.accept)) {
        headers.accept = 'application/json';
    }
    if (options.decompress && is_1.default.undefined(headers['accept-encoding'])) {
        headers['accept-encoding'] = supports_brotli_1.default ? 'gzip, deflate, br' : 'gzip, deflate';
    }
    // Validate URL
    if (options.url.protocol !== 'http:' && options.url.protocol !== 'https:') {
        throw new errors_1.UnsupportedProtocolError(options);
    }
    decodeURI(options.url.toString());
    // Normalize request function
    if (!is_1.default.function_(options.request)) {
        options.request = options.url.protocol === 'https:' ? https.request : http.request;
    }
    // UNIX sockets
    if (options.url.hostname === 'unix') {
        const matches = /(?<socketPath>.+?):(?<path>.+)/.exec(options.url.pathname);
        if ((_a = matches) === null || _a === void 0 ? void 0 : _a.groups) {
            const { socketPath, path } = matches.groups;
            // It's a bug!
            // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
            options = {
                ...options,
                socketPath,
                path,
                host: ''
            };
        }
    }
    if (is_1.default.object(options.agent)) {
        options.agent = options.agent[options.url.protocol.slice(0, -1)] || options.agent;
    }
    if (options.dnsCache) {
        options.lookup = options.dnsCache.lookup;
    }
    /* istanbul ignore next: electron.net is broken */
    // No point in typing process.versions correctly, as
    // `process.version.electron` is used only once, right here.
    if (options.useElectronNet && process.versions.electron) {
        const electron = dynamic_require_1.default(module, 'electron'); // Trick webpack
        options.request = (_b = electron.net.request, (_b !== null && _b !== void 0 ? _b : electron.remote.net.request));
    }
    // Got's `timeout` is an object, http's `timeout` is a number, so they're not compatible.
    delete options.timeout;
    // Set cookies
    if (options.cookieJar) {
        const cookieString = await options.cookieJar.getCookieString(options.url.toString());
        if (is_1.default.nonEmptyString(cookieString)) {
            options.headers.cookie = cookieString;
        }
        else {
            delete options.headers.cookie;
        }
    }
    // `http-cache-semantics` checks this
    delete options.url;
    return options;
};
