"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stream_1 = require("stream");
const is_1 = require("@sindresorhus/is");
function downloadProgress(emitter, downloadBodySize) {
    let downloadedBytes = 0;
    const progressStream = new stream_1.Transform({
        transform(chunk, _encoding, callback) {
            downloadedBytes += chunk.length;
            const percent = downloadBodySize ? downloadedBytes / downloadBodySize : 0;
            // Let `flush()` be responsible for emitting the last event
            if (percent < 1) {
                emitter.emit('downloadProgress', {
                    percent,
                    transferred: downloadedBytes,
                    total: downloadBodySize
                });
            }
            callback(undefined, chunk);
        },
        flush(callback) {
            emitter.emit('downloadProgress', {
                percent: 1,
                transferred: downloadedBytes,
                total: downloadBodySize
            });
            callback();
        }
    });
    return progressStream;
}
exports.downloadProgress = downloadProgress;
function uploadProgress(request, emitter, uploadBodySize) {
    const uploadEventFrequency = 150;
    let uploadedBytes = 0;
    let progressInterval;
    emitter.emit('uploadProgress', {
        percent: 0,
        transferred: 0,
        total: uploadBodySize
    });
    request.once('error', () => {
        clearInterval(progressInterval);
    });
    request.once('abort', () => {
        clearInterval(progressInterval);
    });
    request.once('response', () => {
        clearInterval(progressInterval);
        emitter.emit('uploadProgress', {
            percent: 1,
            transferred: uploadedBytes,
            total: uploadBodySize
        });
    });
    request.once('socket', (socket) => {
        const onSocketConnect = () => {
            progressInterval = setInterval(() => {
                const lastUploadedBytes = uploadedBytes;
                /* istanbul ignore next: future versions of Node may not have this property */
                if (!is_1.default.string(request._header)) {
                    clearInterval(progressInterval);
                    const url = new URL('https://github.com/sindresorhus/got/issues/new');
                    url.searchParams.set('title', '`request._header` is not present');
                    url.searchParams.set('body', 'It causes `uploadProgress` to fail.');
                    console.warn('`request._header` is not present. Please report this as a bug:\n' + url.href);
                    return;
                }
                const headersSize = Buffer.byteLength(request._header);
                uploadedBytes = socket.bytesWritten - headersSize;
                // Don't emit events with unchanged progress and
                // prevent last event from being emitted, because
                // it's emitted when `response` is emitted
                if (uploadedBytes === lastUploadedBytes || uploadedBytes === uploadBodySize) {
                    return;
                }
                emitter.emit('uploadProgress', {
                    percent: uploadBodySize ? uploadedBytes / uploadBodySize : 0,
                    transferred: uploadedBytes,
                    total: uploadBodySize
                });
            }, uploadEventFrequency);
        };
        /* istanbul ignore next: hard to test */
        if (socket.connecting) {
            socket.once('connect', onSocketConnect);
        }
        else if (socket.writable) {
            // The socket is being reused from pool,
            // so the connect event will not be emitted
            onSocketConnect();
        }
    });
}
exports.uploadProgress = uploadProgress;
