"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stream = require("stream");
const EventEmitter = require("events");
const http = require("http");
const CacheableRequest = require("cacheable-request");
const is_1 = require("@sindresorhus/is");
const http_timer_1 = require("@szmarczak/http-timer");
const timed_out_1 = require("./utils/timed-out");
const calculate_retry_delay_1 = require("./calculate-retry-delay");
const get_response_1 = require("./get-response");
const normalize_arguments_1 = require("./normalize-arguments");
const progress_1 = require("./progress");
const errors_1 = require("./errors");
const url_to_options_1 = require("./utils/url-to-options");
const redirectCodes = new Set([300, 301, 302, 303, 304, 307, 308]);
exports.default = (options) => {
    const emitter = new EventEmitter();
    const requestURL = options.url.toString();
    const redirects = [];
    let retryCount = 0;
    let currentRequest;
    let shouldAbort = false;
    const emitError = async (error) => {
        try {
            for (const hook of options.hooks.beforeError) {
                // eslint-disable-next-line no-await-in-loop
                error = await hook(error);
            }
            emitter.emit('error', error);
        }
        catch (error_) {
            emitter.emit('error', error_);
        }
    };
    const get = async () => {
        let httpOptions = await normalize_arguments_1.normalizeRequestArguments(options);
        let timings;
        const handleResponse = async (response) => {
            var _a;
            try {
                /* istanbul ignore next: fixes https://github.com/electron/electron/blob/cbb460d47628a7a146adf4419ed48550a98b2923/lib/browser/api/net.js#L59-L65 */
                if (options.useElectronNet) {
                    response = new Proxy(response, {
                        get: (target, name) => {
                            if (name === 'trailers' || name === 'rawTrailers') {
                                return [];
                            }
                            const value = target[name];
                            return is_1.default.function_(value) ? value.bind(target) : value;
                        }
                    });
                }
                const { statusCode } = response;
                const typedResponse = response;
                // This is intentionally using `||` over `??` so it can also catch empty status message.
                typedResponse.statusMessage = typedResponse.statusMessage || http.STATUS_CODES[statusCode];
                typedResponse.url = options.url.toString();
                typedResponse.requestUrl = requestURL;
                typedResponse.retryCount = retryCount;
                typedResponse.timings = timings;
                typedResponse.redirectUrls = redirects;
                typedResponse.request = { options };
                typedResponse.isFromCache = (_a = typedResponse.fromCache, (_a !== null && _a !== void 0 ? _a : false));
                delete typedResponse.fromCache;
                if (!typedResponse.isFromCache) {
                    // @ts-ignore Node.js typings haven't been updated yet
                    typedResponse.ip = response.socket.remoteAddress;
                }
                const rawCookies = typedResponse.headers['set-cookie'];
                if (Reflect.has(options, 'cookieJar') && rawCookies) {
                    let promises = rawCookies.map((rawCookie) => options.cookieJar.setCookie(rawCookie, typedResponse.url));
                    if (options.ignoreInvalidCookies) {
                        promises = promises.map(p => p.catch(() => { }));
                    }
                    await Promise.all(promises);
                }
                if (options.followRedirect && Reflect.has(typedResponse.headers, 'location') && redirectCodes.has(statusCode)) {
                    typedResponse.resume(); // We're being redirected, we don't care about the response.
                    if (statusCode === 303) {
                        if (options.method !== 'GET' && options.method !== 'HEAD') {
                            // Server responded with "see other", indicating that the resource exists at another location,
                            // and the client should request it from that location via GET or HEAD.
                            options.method = 'GET';
                        }
                        delete options.body;
                        delete options.json;
                        delete options.form;
                    }
                    if (redirects.length >= options.maxRedirects) {
                        throw new errors_1.MaxRedirectsError(typedResponse, options.maxRedirects, options);
                    }
                    // Handles invalid URLs. See https://github.com/sindresorhus/got/issues/604
                    const redirectBuffer = Buffer.from(typedResponse.headers.location, 'binary').toString();
                    const redirectURL = new URL(redirectBuffer, options.url);
                    // Redirecting to a different site, clear cookies.
                    if (redirectURL.hostname !== options.url.hostname) {
                        delete options.headers.cookie;
                    }
                    redirects.push(redirectURL.toString());
                    options.url = redirectURL;
                    for (const hook of options.hooks.beforeRedirect) {
                        // eslint-disable-next-line no-await-in-loop
                        await hook(options, typedResponse);
                    }
                    emitter.emit('redirect', response, options);
                    await get();
                    return;
                }
                get_response_1.default(typedResponse, options, emitter);
            }
            catch (error) {
                emitError(error);
            }
        };
        const handleRequest = (request) => {
            if (shouldAbort) {
                request.abort();
                return;
            }
            currentRequest = request;
            // `request.aborted` is a boolean since v11.0.0: https://github.com/nodejs/node/commit/4b00c4fafaa2ae8c41c1f78823c0feb810ae4723#diff-e3bc37430eb078ccbafe3aa3b570c91a
            // We need to allow `TimedOutTimeoutError` here, because it `stream.pipeline(…)` aborts it automatically.
            const isAborted = () => typeof request.aborted === 'number' || request.aborted === true;
            const onError = (error) => {
                const isTimedOutError = error instanceof timed_out_1.TimeoutError;
                if (!isTimedOutError && isAborted()) {
                    return;
                }
                if (isTimedOutError) {
                    // @ts-ignore TS is dumb.
                    error = new errors_1.TimeoutError(error, timings, options);
                }
                else {
                    error = new errors_1.RequestError(error, options);
                }
                if (emitter.retry(error) === false) {
                    emitError(error);
                }
            };
            const uploadComplete = (error) => {
                if (error) {
                    onError(error);
                    return;
                }
                // No need to attach an error handler here,
                // as `stream.pipeline(…)` doesn't remove this handler
                // to allow stream reuse.
                request.emit('upload-complete');
            };
            request.on('error', onError);
            timings = http_timer_1.default(request); // TODO: Make `@szmarczak/http-timer` set `request.timings` and `response.timings`
            const uploadBodySize = httpOptions.headers['content-length'] ? Number(httpOptions.headers['content-length']) : undefined;
            progress_1.uploadProgress(request, emitter, uploadBodySize);
            timed_out_1.default(request, options.timeout, options.url);
            emitter.emit('request', request);
            if (isAborted()) {
                return;
            }
            try {
                if (options.method === 'POST' || options.method === 'PUT' || options.method === 'PATCH') {
                    if (is_1.default.nodeStream(httpOptions.body)) {
                        // `stream.pipeline(…)` handles `error` for us.
                        request.removeListener('error', onError);
                        stream.pipeline(
                        // @ts-ignore Upgrade `@sindresorhus/is`
                        httpOptions.body, request, uploadComplete);
                    }
                    else {
                        request.end(httpOptions.body, uploadComplete);
                    }
                }
                else {
                    request.end(uploadComplete);
                }
            }
            catch (error) {
                emitError(new errors_1.RequestError(error, options));
            }
        };
        if (options.cache) {
            // `cacheable-request` doesn't support Node 10 API, fallback.
            httpOptions = {
                ...httpOptions,
                ...url_to_options_1.default(options.url)
            };
            const cacheRequest = options.cacheableRequest(httpOptions, handleResponse);
            cacheRequest.once('error', error => {
                if (error instanceof CacheableRequest.RequestError) {
                    emitError(new errors_1.RequestError(error, options));
                }
                else {
                    emitError(new errors_1.CacheError(error, options));
                }
            });
            cacheRequest.once('request', handleRequest);
        }
        else {
            // Catches errors thrown by calling `requestFn(…)`
            try {
                // @ts-ignore 1. TS complains that URLSearchParams is not the same as URLSearchParams.
                //            2. It doesn't notice that `options.timeout` is deleted above.
                handleRequest(httpOptions.request(options.url, httpOptions, handleResponse));
            }
            catch (error) {
                emitError(new errors_1.RequestError(error, options));
            }
        }
    };
    emitter.retry = error => {
        let backoff;
        retryCount++;
        try {
            backoff = options.retry.calculateDelay({
                attemptCount: retryCount,
                retryOptions: options.retry,
                error,
                computedValue: calculate_retry_delay_1.default({
                    attemptCount: retryCount,
                    retryOptions: options.retry,
                    error,
                    computedValue: 0
                })
            });
        }
        catch (error_) {
            emitError(error_);
            return false;
        }
        if (backoff) {
            const retry = async (options) => {
                try {
                    for (const hook of options.hooks.beforeRetry) {
                        // eslint-disable-next-line no-await-in-loop
                        await hook(options, error, retryCount);
                    }
                    await get();
                }
                catch (error_) {
                    emitError(error_);
                }
            };
            setTimeout(retry, backoff, { ...options, forceRefresh: true });
            return true;
        }
        return false;
    };
    emitter.abort = () => {
        if (currentRequest) {
            currentRequest.abort();
        }
        else {
            shouldAbort = true;
        }
    };
    setImmediate(async () => {
        try {
            for (const hook of options.hooks.beforeRequest) {
                // eslint-disable-next-line no-await-in-loop
                await hook(options);
            }
            await get();
        }
        catch (error) {
            emitError(error);
        }
    });
    return emitter;
};
exports.proxyEvents = (proxy, emitter) => {
    const events = [
        'request',
        'redirect',
        'uploadProgress',
        'downloadProgress'
    ];
    for (const event of events) {
        emitter.on(event, (...args) => {
            proxy.emit(event, ...args);
        });
    }
};
